// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * Copyright (C) 2020 VeriSilicon Holdings Co., Ltd.
 *
 * Support manage two slices memory;
 * the memory is in ep side, so the address is in ep pcie axi master view
 * memory space. Every slice have 8GB DDR memory,
 * aggress as follows: the first slice is slice_1, another is slice_2.
 * for each malloc, the minimal memory unit is chunk = 4K
 * for each task, the minimal memory unit is block = 4G/S1_BLOCK_CNT (64M)
 */

#include <linux/module.h>
#include <linux/miscdevice.h>
#include <linux/pci.h>
#include <linux/pagemap.h>

#include "common.h"
#include "memory.h"
#include "transcoder.h"

/* if not used on FPGA, please skip DDR_ECC range.
   change this end address based on actual DDR size.

    0x00000    - 0x80000-1:      512KB memory for micorblaze firmwares loading.
    0x80000    - 0xA0000-1:      128KB share memory from zsp to kernel driver.
    0xA0000    - 0xC0000-1:      128KB share memory from kernel driver to zsp.
    0xC0000    - 0x100000-1:     256KB reserved for future use.
    0x100000   - 0x400000-1:     3M dma link table,
    0x400000   - 0x1800000-1:    20M VCMD,
    0x1800000  - 0x8000000-1:    104M vip external on bar
    0x8000000  - 0x20000000-1:   384M for reserved
    0x20000000 - 0x20800000-1:   8M vip exclusive memory
    0x20800000 - 0xE0000000-1:   3064M for applications
    0xE0000000 - 0x100000000-1:  512M ECC
*/
unsigned long s1_start = 0x20800000UL;
unsigned long s1_end = 0xE0000000UL - 1;

/* if not used on FPGA, please skip DDR_ECC range.
   change this end address based on actual DDR size.

    0x200000000 - 0x2E0000000-1: 3584MB for applications
    0x2E0000000 - 0x300000000:   512M ECC
*/
unsigned long s2_start = 0x200000000UL;
unsigned long s2_end = 0x2E0000000UL - 1;

module_param(s1_start, ulong, 0644);
MODULE_PARM_DESC(s1_start, "slice1 ddr start address; default is 0x18000000.");

module_param(s1_end, ulong, 0644);
MODULE_PARM_DESC(s1_end, "slice1 ddr end address; default is 0x1c0000000.");

module_param(s2_start, ulong, 0644);
MODULE_PARM_DESC(s2_start, "slice2 ddr start address; default is 0x218000000.");

module_param(s2_end, ulong, 0644);
MODULE_PARM_DESC(s2_end, "slice2 ddr end address; default is 0x3c0000000.");



/* taskid status */
#define ID_BUSY 0x1
#define ID_FREE 0x0

#define S1_BLK_SIZE (((s1_end - s1_start) / S1_BLOCK_CNT) & (~(CHUNK_SIZE - 1)))
#define S2_BLK_SIZE (((s2_end - s2_start) / S2_BLOCK_CNT) & (~(CHUNK_SIZE - 1)))
#define MAX_BLK_SIZE (S2_BLK_SIZE > S1_BLK_SIZE ? S2_BLK_SIZE : S1_BLK_SIZE)

/* memory chunk information */
struct hlina_chunk {
	u64 bus_add; /* this chunk start physical address */
	u32 ck_rsv;  /* indicate this chunk whether be used or not */
};

/* memory block information */
struct mem_block {
	u32 index; /* block index */
	u64 start_addr; /*  the block start address; */
	int task_id; /* this block be used by task_id */
	int fd; /* this block be used by fd */
	u32 block_size; /* block size; */
	long long chunks; /* total chunk count */
	long long free_chunks; /* the chunks not used */
	struct hlina_chunk *chk_inf;
	int alloc_blk_flag; /* allocation flags */
};

/* get a not used task id, maximum support MAX_TASK_NUM task, so total count of
 * task id is MAX_TASK_NUM, from 1 to MAX_TASK_NUM.
 */
int get_task_id(struct memory_t *tmem, struct file *filp)
{
	int i;

	spin_lock(&tmem->taskid_lock);
	for (i = MIN_TASK_ID; i <= MAX_TASK_NUM; i++) {
		if (tmem->id_st[i] == ID_FREE) {
			tmem->id_st[i] = ID_BUSY;
			tmem->id_filp[i] = filp;
			break;
		}
	}
	spin_unlock(&tmem->taskid_lock);

	return (i <= MAX_TASK_NUM) ? i : -1;
}

/* if free a task id, will free all corresponding memory. */
void free_task_id(struct memory_t *tmem, int id)
{
	int i, j;
	struct mem_block *sbk;
	int mem_leak_flag_1 = 0;
	int mem_leak_flag_2 = 0;

	if (mutex_lock_interruptible(&tmem->mem_mutex_ep))
		return;

	spin_lock(&tmem->taskid_lock);
	tmem->id_st[id] = ID_FREE;
	tmem->id_filp[id] = NULL;

	for (i = 0; i < S1_BLOCK_CNT; i++) {
		if (tmem->mem_s1_bk[i].task_id == id) {
			sbk = &(tmem->mem_s1_bk[i]);
			for (j = 0; j < sbk->chunks; j++) {
				if (sbk->chk_inf[j].ck_rsv) {
					mem_leak_flag_1++;
					tmem->s1_rev_size -=
						sbk->chk_inf[j].ck_rsv *
						CHUNK_SIZE;
					sbk->chk_inf[j].ck_rsv = 0;
					sn_pri(tmem->tdev, SN_DBG,
					       "mem: mem leak,s1,blk:%d,id:%d,add=0x%llx\n",
					       i, id, sbk->chk_inf[j].bus_add);
				}
			}
			sbk->task_id = 0;
			sbk->fd = 0;
			sbk->alloc_blk_flag = 0;
		}
	}
	for (i = 0; i < S2_BLOCK_CNT; i++) {
		if (tmem->mem_s2_bk[i].task_id == id) {
			sbk = &(tmem->mem_s2_bk[i]);
			for (j = 0; j < sbk->chunks; j++) {
				if (sbk->chk_inf[j].ck_rsv) {
					mem_leak_flag_2++;
					tmem->s2_rev_size -=
						sbk->chk_inf[j].ck_rsv *
						CHUNK_SIZE;
					sbk->chk_inf[j].ck_rsv = 0;
					sn_pri(tmem->tdev, SN_DBG,
					       "mem: mem leak,s2,blk:%d,id:%d,add=0x%llx\n",
					       i, id, sbk->chk_inf[j].bus_add);
				}
			}
			sbk->task_id = 0;
			sbk->fd = 0;
			sbk->alloc_blk_flag = 0;
		}
	}
	if (mem_leak_flag_1)
		sn_pri(tmem->tdev, SN_DBG, "mem: mem leak in slice_1\n");
	if (mem_leak_flag_2)
		sn_pri(tmem->tdev, SN_DBG, "mem: mem leak in slice_2\n");

	spin_unlock(&tmem->taskid_lock);
	mutex_unlock(&tmem->mem_mutex_ep);
}

void sn_mem_close(struct sn_tranx_t *tdev, struct file *filp)
{
	int id;
	struct memory_t *tmem = tdev->modules[SN_MODULE_MEMORY];

	for (id = MIN_TASK_ID; id <= MAX_TASK_NUM; id++) {
		if (tmem->id_filp[id] == filp) {
			sn_pri(tdev, SN_DBG, "%s id:%d\n", __func__, id);
			free_task_id(tmem, id);
		}
	}
}

static int get_blocks(int slice, unsigned long *busaddr,
		      unsigned int size, int task_id, int fd, struct memory_t *tmem)
{
	/* suppose all the block size is same */
	struct mem_block *mb;
	unsigned int block_size = 0;
	unsigned int blk_num = 0, total_blk = 0;
	unsigned int i = 0;
	int item = -1;

	if (slice == SLICE1_INDEX) {
		mb = tmem->mem_s1_bk;
		total_blk = S1_BLOCK_CNT;
	} else {
		mb = tmem->mem_s2_bk;
		total_blk = S2_BLOCK_CNT;
	}

	block_size = mb[0].block_size;
	blk_num = size / block_size + (size % block_size > 0 ? 1 : 0);

	for (i = 0; i < (total_blk - blk_num); i++) {
		if (mb[i].free_chunks == mb[i].chunks) {
			int j = 0;

			while (++j < blk_num) {
				if (mb[i + j].free_chunks != mb[i + j].chunks) {
					break;
				}
			}

			if (j == blk_num) { //found
				item = i;
				break;
			} else {
				i += j;  //not found, continue
			}
		}
	}

	if (item != -1) {
		//set these blocks occupied
		for (i = 0; i < blk_num; i++) {
			mb[item + i].task_id = task_id;
			mb[item + i].chk_inf[0].ck_rsv = mb[item + i].chunks;
			mb[item + i].free_chunks = 0;
			mb[item + i].fd = fd;
			mb[item + i].alloc_blk_flag = item + 1; //mark these blocks belong to same allocation

			if (slice == SLICE1_INDEX)
				tmem->s1_rev_size += mb[item + i].block_size;
			else
				tmem->s2_rev_size += mb[item + i].block_size;
		}

		*busaddr = mb[item].start_addr;
		return 0;
	}
	return -EFAULT;
}

static int free_blocks(struct mem_block *mb, unsigned int bi, int blk_cnt, unsigned long busaddr,
		       unsigned int size, int task_id, struct memory_t *tmem)
{
	int slice = 0, i = 0, j = 0;
	int ret = -EAGAIN, flag_check = mb[bi].alloc_blk_flag;
	unsigned int free_size = 0;

	/* need free chunks */
	if (mb[bi].start_addr != busaddr) {
		return -EAGAIN;
	}

	for (i = bi; i < blk_cnt; i++) {
		if (mb[i].alloc_blk_flag && mb[i].task_id == task_id) {
			//avlid over free
			if (flag_check != mb[i].alloc_blk_flag)
				break;

			if (slice == SLICE1_INDEX)
				tmem->s1_rev_size -= mb[i].block_size;
			else
				tmem->s2_rev_size -= mb[i].block_size;

			mb[i].free_chunks = mb[bi].chunks;
			for (j = 0; j < mb[i].chunks; j++)
				mb[i].chk_inf[j].ck_rsv = 0;

			mb[i].task_id = 0;
			mb[i].fd = 0;
			mb[i].alloc_blk_flag = 0;
			ret = 0;
			free_size += mb[i].block_size;
		}
	}
	sn_pri(tmem->tdev, SN_DBG, "mem: %s, the size %d is freed, total free %d.\n",
		__func__, size, free_size);

	return ret;
}

/* Cycle through the buffers we have, give the first free one */
static int get_chunks(int slice, int block_index, unsigned long *busaddr,
		      unsigned int size, int task_id, struct memory_t *tmem)
{
	int i, j;
	int ret = 0;
	int bi = block_index;
	struct mem_block *mb;
	unsigned int skip_chunks = 0;
	/* calculate how many chunks we need;round up to chunk boundary */
	long alloc_chunks = (size + CHUNK_SIZE - 1) / CHUNK_SIZE;

	if (slice > SLICE2_INDEX) {
		sn_pri(tmem->tdev, SN_ERR, "mem: %s, slice id:%d is error\n",
		       __func__, slice);
		return -EFAULT;
	}

	if (slice == SLICE1_INDEX)
		mb = tmem->mem_s1_bk;
	else
		mb = tmem->mem_s2_bk;

	*busaddr = 0;
	/* run through the chunk table */
	for (i = 0; i < mb[bi].chunks;) {
		skip_chunks = 0;
		/* if this chunk is available */
		if (!mb[bi].chk_inf[i].ck_rsv) {
			/* check that there is enough memory left */
			if (i + alloc_chunks > mb[bi].chunks) {
				sn_pri(tmem->tdev, SN_DBG,
		       			"mem: debug info:get chunks has not enough mem\n");
				break;
			}

			/*
			 * check that there is enough consecutive
			 * chunks available.
			 */
			for (j = i; j < i + alloc_chunks; j++) {
				if (mb[bi].chk_inf[j].ck_rsv) {
					skip_chunks = 1;
					/* skip the used chunks */
					i = j + mb[bi].chk_inf[j].ck_rsv;
					break;
				}
			}

			/* if enough free memory found */
			if (!skip_chunks) {
				/* task chunk allocation info */
				sn_pri(tmem->tdev, SN_DBG, "task_id[%d]: alloc chunks %d[%d][%d], %d, %d.\n", task_id, slice, bi, i, alloc_chunks, mb[bi].free_chunks);

				*busaddr = mb[bi].chk_inf[i].bus_add;
				mb[bi].chk_inf[i].ck_rsv = alloc_chunks;
				mb[bi].free_chunks -= alloc_chunks;

				if (slice == SLICE1_INDEX)
					tmem->s1_rev_size +=
						alloc_chunks * CHUNK_SIZE;
				else
					tmem->s2_rev_size +=
						alloc_chunks * CHUNK_SIZE;

				break;
			}
		} else {
			/* skip the used chunks */
			i += mb[bi].chk_inf[i].ck_rsv;
		}
	}

	if (*busaddr == 0) {
		sn_pri(tmem->tdev, SN_DBG,
		       "mem: debug info:get chunks failed in s:%d,block:%d,size:0x%x,task_id:%d\n",
		       slice, bi, size, task_id);
		ret = -EFAULT;
	} else
		sn_pri(tmem->tdev, SN_DBG,
		       "mem: got mem in s:%d block:%d,addr:0x%lx size:0x%x,task_id:%d\n",
		       slice, bi, *busaddr, size, task_id);

	return ret;
}

/* allocate memory from a new block */
int alloc_in_new_block(int slice, unsigned long *addr, unsigned int size,
		       int task_id, int fd, struct memory_t *tmem)
{
	int i, j, blk_cnt;
	int ret = 0;
	struct mem_block *mem_block;

	if (slice == SLICE1_INDEX) {
		mem_block = tmem->mem_s1_bk;
		blk_cnt = S1_BLOCK_CNT;
	} else if (slice == SLICE2_INDEX) {
		mem_block = tmem->mem_s2_bk;
		blk_cnt = S2_BLOCK_CNT;
	} else {
		sn_pri(tmem->tdev, SN_ERR, "mem: %s, slice id:%d is error\n",
		       __func__, slice);
		return -EFAULT;
	}

	*addr = 0;
	/* traversal all blocks, get a new block. */
	for (i = 0; i < blk_cnt; i++) {
		if (mem_block[i].task_id == 0) {
			ret = get_chunks(slice, i, addr, size, task_id, tmem);
			if (ret) {
				sn_pri(tmem->tdev, SN_ERR,
				       "mem: the size:0x%x is very large, more than slice:%d block:%d size, task_id=%d\n",
				       size, slice, i, task_id);
				for (j = 0; j < mem_block[i].chunks; j++) {
					if (mem_block[i].chk_inf[j].ck_rsv)
						sn_pri(tmem->tdev, SN_ERR,
						       "mem: j:%d reserve_cnt:%d\n",
						       j,
						       mem_block[i].chk_inf[j].ck_rsv);
				}
			} else {
				mem_block[i].task_id = task_id;
				mem_block[i].fd = fd;
				sn_pri(tmem->tdev, SN_DBG,
				       "Allocate mem in slice%d block%i addr=%lx, start=%llx, task_id=%d, pid=%d\n",
				       slice, i, *addr, mem_block[i].start_addr,
				       mem_block[i].task_id, fd);
				break;
			}
		}
	}

	/*  not has valid block, alloc failed. */
	if (i == blk_cnt) {
		sn_pri(tmem->tdev, SN_DBG,
		       "mem: %s, not has valid block in slice%d.\n", __func__,
		       slice);
		return -EFAULT;
	}

	return ret;
}

/*
 * get memory from a slice.
 * @s: slice id.
 * @id: task id
 * @check: check:1, only get memory from used block
 *         check:0, if can't get valid memory in used block, allocate from
 *                  a new block.
 */
static int alloc_mem_in_slice(int s, unsigned long *addr, unsigned int size,
			      int id, int fd, struct memory_t *tmem, int check)
{
	int i, blk_cnt;
	int ret = 0;
	struct mem_block *mem_block;
	unsigned int blk_max_size = 0;

	if (s == SLICE1_INDEX) {
		mem_block = tmem->mem_s1_bk;
		blk_cnt = S1_BLOCK_CNT;
		blk_max_size = S1_BLK_SIZE;
	} else if (s == SLICE2_INDEX) {
		mem_block = tmem->mem_s2_bk;
		blk_cnt = S2_BLOCK_CNT;
		blk_max_size = S2_BLK_SIZE;
	} else {
		sn_pri(tmem->tdev, SN_ERR,
		       "mem: slice id:%d is error in alloc mem in slice\n", s);
		return -EFAULT;
	}
	*addr = 0;

	if (size > blk_max_size) {
		sn_pri(tmem->tdev, SN_INF, "mem: get mem in block allocator size 0x%x.\n", size);
		return get_blocks(s, addr, size, id, fd, tmem);
	}

	/* traversal all blocks,check itself block if it has valid memory */
	for (i = 0; i < blk_cnt; i++) {
		if (mem_block[i].task_id == id) {
			ret = get_chunks(s, i, addr, size, id, tmem);
			if (!ret)
				break;
		}
	}

	/* not has valid memory in itself block */
	if (i == blk_cnt) {
		if (check == 1)
			ret = -EFAULT;
		else
			/* get a new block; */
			ret = alloc_in_new_block(s, addr, size, id, fd, tmem);
	}

	return ret;
}

/* allocate memory from its own used block */
static int alloc_in_used_block(unsigned long *busaddr, unsigned int size,
			       int task_id, int fd, struct memory_t *tmem)
{
	int ret;

	if (tmem->s1_rev_size <= tmem->s2_rev_size) {
		/* check slice1 if there is valid used block */
		ret = alloc_mem_in_slice(SLICE1_INDEX, busaddr, size, task_id,
					 fd, tmem, 1);
		/* s1 not has valid used blocks for this id, check s2. */
		if (ret) {
			/* check slice2 if it has valid used block. */
			ret = alloc_mem_in_slice(SLICE2_INDEX, busaddr, size,
						 task_id, fd, tmem, 1);
		}
	} else {
		/* check slice1 if it has valid used block */
		ret = alloc_mem_in_slice(SLICE2_INDEX, busaddr, size, task_id,
					 fd, tmem, 1);
		/* s2 not has valid used blocks for this id, check s1. */
		if (ret) {
			/* check s1 if it has valid used block. */
			ret = alloc_mem_in_slice(SLICE1_INDEX, busaddr, size,
						 task_id, fd, tmem, 1);
		}
	}

	return ret;
}

/*
 * allocate memory in ep side.
 * tmem->tdev->mem_alloc_method is the allocation strategy.
 *	1: only use slice_1
 *	2: only use_slice_2
 *	3: balance s1 and s2 by used size.
 */
int alloc_mem_ep(unsigned long *busaddr, unsigned int size, int task_id, int fd,
		 struct memory_t *tmem)
{
	int ret = 0;
	*busaddr = 0;

	if (WARN_ON(tmem->tdev->mem_alloc_method > ALLOCATE_METHOD_S1S2)) {
		sn_pri(tmem->tdev, SN_ERR, "mem: %s, method:%d error\n",
		       __func__);
		return -EFAULT;
	}

	if (mutex_lock_interruptible(&tmem->mem_mutex_ep))
		return -ERESTARTSYS;

	/* alloc memory from slice 1  */
	if (tmem->tdev->mem_alloc_method == ALLOCATE_METHOD_S1)
		ret = alloc_mem_in_slice(SLICE1_INDEX, busaddr, size, task_id,
					 fd, tmem, 0);

	/* alloc memory from slice 2 */
	else if (tmem->tdev->mem_alloc_method == ALLOCATE_METHOD_S2)
		ret = alloc_mem_in_slice(SLICE2_INDEX, busaddr, size, task_id,
					 fd, tmem, 0);

	/* alloc memory from slice1 and slice2, a simple balance */
	else {
		if (alloc_in_used_block(busaddr, size, task_id, fd, tmem)) {
			if (tmem->s1_rev_size <= tmem->s2_rev_size) {
				/* first alloc from s1 */
				ret = alloc_in_new_block(SLICE1_INDEX, busaddr,
							 size, task_id, fd, tmem);
				if (ret)
					ret = alloc_in_new_block(SLICE2_INDEX,
								 busaddr, size,task_id, fd, tmem);
			} else {
				/* first alloc from s2 */
				ret = alloc_in_new_block(SLICE2_INDEX, busaddr,
							 size, task_id, fd, tmem);
				if (ret)
					ret = alloc_in_new_block(SLICE1_INDEX,
								 busaddr, size, task_id, fd, tmem);
			}
		}
	}

	if (!*busaddr) {
		sn_pri(tmem->tdev, SN_DBG, "mem: alloc ep mem failed\n");
		ret = -EFAULT;
	}

	mutex_unlock(&tmem->mem_mutex_ep);

	return ret;
}

/*
 * when free memory, will check the block which did this memory used
 * if this block is not used, free it.
 */
int free_mem_ep(unsigned long busaddr, unsigned int size, int task_id,
		struct memory_t *tmem)
{
	int i, ret = 0;
	int slice, bi, blk_cnt = -1; /* bi:block index */
	struct mem_block *mb; /* memory block */
	unsigned long tmp;
	u32 rsv_size;

	if (mutex_lock_interruptible(&tmem->mem_mutex_ep))
		return -ERESTARTSYS;
	/* Find the slice it belongs to */
	if ((busaddr >= s1_start) && (busaddr < s1_end)) {
		slice = SLICE1_INDEX;
		mb = tmem->mem_s1_bk;
		blk_cnt = S1_BLOCK_CNT;

		/* Find the block it belongs to */
		for (i = 0; i < blk_cnt; i++) {
			tmp = mb[i].start_addr + mb[i].block_size;
			if ((busaddr >= mb[i].start_addr) && (busaddr < tmp)) {
				bi = i;
				break;
			}
		}
	} else if ((busaddr >= s2_start) && (busaddr < s2_end)) {
		slice = SLICE2_INDEX;
		mb = tmem->mem_s2_bk;
		blk_cnt = S2_BLOCK_CNT;

		for (i = 0; i < blk_cnt; i++) {
			tmp = mb[i].start_addr + mb[i].block_size;
			if ((busaddr >= mb[i].start_addr) && (busaddr < tmp)) {
				bi = i;
				break;
			}
		}
	} else {
		sn_pri(tmem->tdev, SN_ERR, "mem: busaddr:0x%lx is invalid\n",
		       busaddr);
		ret = -EFAULT;
		goto err;
	}
	if (bi == -1) {
		sn_pri(tmem->tdev, SN_ERR,
		       "mem: busaddr:0x%lx is invalid, can't find the block\n",
		       busaddr);
		ret = -EFAULT;
		goto err;
	}

	if (mb[bi].task_id != task_id) {
		sn_pri(tmem->tdev, SN_ERR,
		       "mem: %s Wrong task_id:%d on block:%d,expected id is %d\n",
		       __func__, task_id, bi, mb[bi].task_id);
		sn_pri(tmem->tdev, SN_ERR,
		       "mem: %s Owner mismatch while freeing memory!\n",
		       __func__);
		ret = -EFAULT;
		goto err;
	}

	/* the size is not check if wrong size give */
	if (mb[bi].alloc_blk_flag) {
		if (0 == free_blocks(mb, bi, blk_cnt, busaddr, size, task_id, tmem)) {
			/* success free, just skip the chunk free */
			sn_pri(tmem->tdev, SN_INF, "mem: busaddr:0x%lx free in block free.\n", busaddr);
			ret = 0;
			goto err;
		}
	}

	sn_pri(tmem->tdev, SN_DBG,
	       "mem: %s, busaddr:0x%lx, size:0x%x, in slice:%d block:%d.\n",
	       __func__, busaddr, size, slice, bi);

	for (i = 0; i < mb[bi].chunks; i++) {
		if (mb[bi].chk_inf[i].bus_add == busaddr) {
			rsv_size = mb[bi].chk_inf[i].ck_rsv * CHUNK_SIZE;
			if (slice == SLICE1_INDEX)
				tmem->s1_rev_size -= rsv_size;
			else
				tmem->s2_rev_size -= rsv_size;

			/* task chunk free info */
			sn_pri(tmem->tdev, SN_DBG, "task_id %d: free chunks %d[%d][%d], %d.\n", task_id, slice, bi, i, mb[bi].chk_inf[i].ck_rsv);

			mb[bi].free_chunks += mb[bi].chk_inf[i].ck_rsv;
			mb[bi].chk_inf[i].ck_rsv = 0;

			if (mb[bi].free_chunks == mb[bi].chunks) {
				sn_pri(tmem->tdev, SN_DBG,
				       "mem: %s block:%d is free in slice:%d\n",
				       __func__, bi, slice);
				mb[bi].task_id = 0;
				mb[bi].fd = 0;
			}
			break;
		}
	}

	if (i == mb[bi].chunks) {
		sn_pri(tmem->tdev, SN_ERR,
		       "mem: %s can't find the chunk,busaddr:0x%lx size:0x%x\n",
		       __func__, busaddr, size);
		ret = -EFAULT;
		goto err;
	}

err:
	mutex_unlock(&tmem->mem_mutex_ep);
	return ret;
}

/* Compute memory utilization */
static void mem_usage(struct sn_tranx_t *tdev, struct mem_used_info *info)
{
	int i;
	u32 s1_used = 0, s2_used = 0;
	struct memory_t *tmem = tdev->modules[SN_MODULE_MEMORY];

	info->s1_blk_used = 0;
	info->s2_blk_used = 0;
	for (i = 0; i < S1_BLOCK_CNT; i++) {
		if (tmem->mem_s1_bk[i].task_id) {
			s1_used += tmem->mem_s1_bk[i].block_size;
			info->s1_blk_used++;
		}
	}
	for (i = 0; i < S2_BLOCK_CNT; i++) {
		if (tmem->mem_s2_bk[i].task_id) {
			s2_used += tmem->mem_s2_bk[i].block_size;
			info->s2_blk_used++;
		}
	}

	info->s1_used = s1_used / 1024 / 1024;
	info->s1_free = (s1_end - s1_start - s1_used) / 1024 / 1024;
	info->s2_used = s2_used / 1024 / 1024;
	info->s2_free = (s2_end - s2_start - s2_used) / 1024 / 1024;
}

/* Check the owner of the input address */
int get_memory_owner(struct sn_tranx_t *tdev, u64 addr)
{
	int i;
	struct memory_t *tmem = tdev->modules[SN_MODULE_MEMORY];

	for (i = 0; i < S1_BLOCK_CNT; i++) {
		if (addr >= tmem->mem_s1_bk[i].start_addr &&
		    addr < tmem->mem_s1_bk[i].start_addr +
				    tmem->mem_s1_bk[i].block_size) {
			return tmem->mem_s1_bk[i].fd;
		}
	}

	for (i = 0; i < S2_BLOCK_CNT; i++) {
		if (addr >= tmem->mem_s2_bk[i].start_addr &&
		    addr < tmem->mem_s2_bk[i].start_addr +
				    tmem->mem_s2_bk[i].block_size) {
			return tmem->mem_s2_bk[i].fd;
		}
	}
	return -1;
}

/* display  memory status */
static ssize_t mem_info_show(struct device *dev, struct device_attribute *attr,
			     char *buf)
{
	int pos = 0;
	struct sn_misc_tdev *mtdev = dev_get_drvdata(dev);
	struct sn_tranx_t *tdev = mtdev->tdev;
	struct memory_t *tmem = tdev->modules[SN_MODULE_MEMORY];
	struct mem_block *s1_bk = tmem->mem_s1_bk;
	struct mem_block *s2_bk = tmem->mem_s2_bk;
	u32 max_len = 10, i, j;
	u32 cnt = 0, remainder = 0;
	struct mem_used_info info;
	u32 s1_total = (s1_end - s1_start) / 1024 / 1024;
	u32 s2_total = (s2_end - s2_start) / 1024 / 1024;

	mem_usage(tdev, &info);
	pos += sprintf(buf + pos,
		       "S1:  %d MB used   %d MB free   %d.%02d%% used\n",
		       info.s1_used, info.s1_free,
		       (info.s1_used * 10000) / (s1_total) / 100,
		       (info.s1_used * 10000) / (s1_total) % 100);
	pos += sprintf(buf + pos,
		       "S2:  %d MB used   %d MB free   %d.%02d%% used\n",
		       info.s2_used, info.s2_free,
		       (info.s2_used * 10000) / (s2_total) / 100,
		       (info.s2_used * 10000) / (s2_total) % 100);

	// print S1:
	pos += sprintf(buf + pos, "\nS1 BK :");
	for (i = 1; i <= max_len; i++)
		pos += sprintf(buf + pos, "  %3d", i);
	pos += sprintf(buf + pos, "\n");

	cnt = S1_BLOCK_CNT / max_len;
	for (j = 0; j < cnt; j++) {
		pos += sprintf(buf + pos, "   %3d:", j * max_len);
		for (i = 0; i < max_len; i++)
			pos += sprintf(buf + pos, "  %3d",
				       s1_bk[j * max_len + i].task_id);
		pos += sprintf(buf + pos, "\n");
	}

	remainder = S1_BLOCK_CNT % max_len;
	if (remainder) {
		pos += sprintf(buf + pos, "   %3d:", cnt * max_len);
		for (i = 0; i < remainder; i++)
			pos += sprintf(buf + pos, "  %3d",
						s1_bk[cnt * max_len + i].task_id);
		pos += sprintf(buf + pos, "\n");
	}

	// print S2:
	pos += sprintf(buf + pos, "\nS2 BK :");
	for (i = 1; i <= max_len; i++)
		pos += sprintf(buf + pos, "  %3d", i);
	pos += sprintf(buf + pos, "\n");

	cnt = S2_BLOCK_CNT / max_len;
	for (j = 0; j < cnt; j++) {
		pos += sprintf(buf + pos, "   %3d:", j * max_len);
		for (i = 0; i < max_len; i++)
			pos += sprintf(buf + pos, "  %3d",
				       s2_bk[j * max_len + i].task_id);
		pos += sprintf(buf + pos, "\n");
	}

	remainder = S2_BLOCK_CNT % max_len;
	if (remainder) {
		pos += sprintf(buf + pos, "   %3d:", cnt * max_len);
		for (i = 0; i < remainder; i++)
			pos += sprintf(buf + pos, "  %3d",
						s2_bk[cnt * max_len + i].task_id);
		pos += sprintf(buf + pos, "\n");
	}
	return pos;
}

static DEVICE_ATTR_RO(mem_info);

static struct attribute *transmem_sysfs_entries[] = { &dev_attr_mem_info.attr,
						      NULL };

static struct attribute_group transmem_attribute_group = {
	.name = NULL,
	.attrs = transmem_sysfs_entries,
};

static int check_task_id(struct memory_t *tmem, int task_id)
{
	if ((task_id > MAX_TASK_NUM) || (task_id < MIN_TASK_ID) ||
	    (tmem->id_st[task_id] == ID_FREE)) {
		return -EFAULT;
	} else
		return 0;
}

int sn_mem_get_method(struct sn_tranx_t *tdev)
{
	int ret = 0;
	/* 1:use slice_1;   2:use_slice_2;    3:balance s1 and s2 */
	if (tdev->pf_vf_mode == TWO_VF_MODE) {
		if (tdev->vf_index == VF2_INDEX) {
			ret = ALLOCATE_METHOD_S2;
		}
		else if (tdev->vf_index == VF1_INDEX) {
			ret = ALLOCATE_METHOD_S1;
		}
	}
	else {
		ret = ALLOCATE_METHOD_S1S2;
	}
	return ret;
}

int sn_mem_init(struct sn_tranx_t *tdev)
{
	int ret = -EFAULT;
	u32 i, j;
	u64 blk_size;
	void *block;
	void *hlina_chunks;
	struct mem_block *mem_s1_bk;
	struct mem_block *mem_s2_bk;
	u64 tmp_addr;
	struct memory_t *tmem;

	if ((s1_end <= s1_start) || (s2_end <= s2_start)) {
		sn_pri(tdev, SN_ERR,
		       "mem: please check memory start address and end address.\n");
		goto out;
	}

	tdev->mem_alloc_method = sn_mem_get_method(tdev);
	tmem = kzalloc(sizeof(struct memory_t), GFP_KERNEL);
	if (!tmem) {
		sn_pri(tdev, SN_ERR, "mem: alloc tmem error\n");
		goto out;
	}
	tdev->modules[SN_MODULE_MEMORY] = tmem;
	tmem->tdev = tdev;

	block = kzalloc(sizeof(struct mem_block) *
				(S1_BLOCK_CNT + S2_BLOCK_CNT),
			GFP_KERNEL);
	if (!block) {
		sn_pri(tdev, SN_ERR, "mem: alloc block error\n");
		goto out_free_dev;
	}

	hlina_chunks = vzalloc((s1_end - s1_start + s2_end - s2_start) /
			       CHUNK_SIZE * sizeof(struct hlina_chunk));
	if (!hlina_chunks) {
		sn_pri(tdev, SN_ERR, "mem: alloc hlina_chunks failed\n");
		goto out_free_slice;
	}

	mem_s1_bk = block;
	blk_size = ((s1_end - s1_start) / S1_BLOCK_CNT) & (~(CHUNK_SIZE - 1));
	sn_pri(tdev, SN_DBG,
	       "memory : s1_start:0x%08lx s1_size:0x%lx block_cnt:%d, block_size:0x%lx,chunks:%lld.\n",
	       s1_start, s1_end - s1_start, S1_BLOCK_CNT, blk_size,
	       blk_size * S1_BLOCK_CNT / CHUNK_SIZE);
	for (i = 0; i < S1_BLOCK_CNT; i++) {
		mem_s1_bk[i].index = i;
		mem_s1_bk[i].task_id = 0;
		mem_s1_bk[i].fd = 0;
		mem_s1_bk[i].block_size = blk_size;
		mem_s1_bk[i].start_addr = s1_start + blk_size * i;
		mem_s1_bk[i].chunks = mem_s1_bk[i].block_size / CHUNK_SIZE;
		mem_s1_bk[i].free_chunks = mem_s1_bk[i].chunks;
		mem_s1_bk[i].chk_inf = hlina_chunks;
		mem_s1_bk[i].alloc_blk_flag = 0;
		hlina_chunks +=
			sizeof(struct hlina_chunk) * mem_s1_bk[i].chunks;
		tmp_addr = mem_s1_bk[i].start_addr;
		for (j = 0; j < mem_s1_bk[i].chunks; j++) {
			mem_s1_bk[i].chk_inf[j].bus_add = tmp_addr;
			mem_s1_bk[i].chk_inf[j].ck_rsv = 0;
			tmp_addr += CHUNK_SIZE;
		}
	}

	mem_s2_bk = block + sizeof(struct mem_block) * S1_BLOCK_CNT;
	blk_size = ((s2_end - s2_start) / S2_BLOCK_CNT) & (~(CHUNK_SIZE - 1));
	sn_pri(tdev, SN_DBG,
	       "memory : s2_start:0x%08lx s2_size:0x%lx block_cnt:%d, block_size:0x%lx,chunks:%ld.\n",
	       s2_start, s2_end - s2_start, S2_BLOCK_CNT, blk_size,
	       blk_size * S1_BLOCK_CNT / CHUNK_SIZE);
	for (i = 0; i < S2_BLOCK_CNT; i++) {
		mem_s2_bk[i].index = i;
		mem_s2_bk[i].task_id = 0;
		mem_s2_bk[i].fd = 0;
		mem_s2_bk[i].block_size = blk_size;
		mem_s2_bk[i].start_addr = s2_start + blk_size * i;
		mem_s2_bk[i].chunks = mem_s2_bk[i].block_size / CHUNK_SIZE;
		mem_s2_bk[i].free_chunks = mem_s2_bk[i].chunks;
		mem_s2_bk[i].chk_inf = hlina_chunks;
		mem_s2_bk[i].alloc_blk_flag = 0;
		hlina_chunks +=
			sizeof(struct hlina_chunk) * mem_s2_bk[i].chunks;
		tmp_addr = mem_s2_bk[i].start_addr;
		for (j = 0; j < mem_s2_bk[i].chunks; j++) {
			mem_s2_bk[i].chk_inf[j].bus_add = tmp_addr;
			mem_s2_bk[i].chk_inf[j].ck_rsv = 0;
			tmp_addr += CHUNK_SIZE;
		}
	}

	tmem->mem_s1_bk = mem_s1_bk;
	tmem->mem_s2_bk = mem_s2_bk;
	tmem->s1_rev_size = 0;
	tmem->s2_rev_size = 0;

	mutex_init(&tmem->mem_mutex_ep);
	spin_lock_init(&tmem->taskid_lock);

	ret = sysfs_create_group(&tdev->misc_dev->this_device->kobj,
				 &transmem_attribute_group);
	if (ret) {
		sn_pri(tdev, SN_ERR, "mem: failed to create sysfs entry\n");
		goto out_free_chunk;
	}

	sn_pri(tdev, SN_INF, "mem: module initialization done\n");
	return 0;

out_free_chunk:
	vfree(hlina_chunks);
out_free_slice:
	kfree(block);
out_free_dev:
	kfree(tmem);
out:
	sn_pri(tdev, SN_ERR, "mem: module initialize failed.\n");
	return ret;
}

void sn_mem_release(struct sn_tranx_t *tdev)
{
	struct memory_t *tmem = tdev->modules[SN_MODULE_MEMORY];

	mutex_destroy(&tmem->mem_mutex_ep);

	vfree(tmem->mem_s1_bk[0].chk_inf);
	kfree(tmem->mem_s1_bk);

	sysfs_remove_group(&tdev->misc_dev->this_device->kobj,
			   &transmem_attribute_group);

	kfree(tmem);
	sn_pri(tdev, SN_DBG, "mem: remove module done.\n");
}

long sn_mem_ioctl(struct file *filp, unsigned int cmd, unsigned long arg,
		  struct sn_tranx_t *tdev)
{
	int ret = 0;
	int id;
	struct mem_used_info info;
	struct mem_info memp;
	unsigned long addr;
	void __user *argp = (void __user *)arg;
	struct memory_t *tmem = tdev->modules[SN_MODULE_MEMORY];

	switch (cmd) {
	case SN_TRANX_MEM_GET_UTIL:
		mem_usage(tdev, &info);
		if (__copy_to_user(argp, &info, sizeof(info))) {
			sn_pri(tdev, SN_ERR,
			       "mem: get mem usage - copy_to_user failed.\n");
			ret = -EFAULT;
		}
		break;
	case SN_TRANX_MEM_ALLOC:
		if (copy_from_user(&memp, argp, sizeof(memp))) {
			sn_pri(tdev, SN_ERR,
			       "mem: get mem - copy_from_user failed.\n");
			return -EFAULT;
		}
		if (check_task_id(tmem, memp.task_id)) {
			sn_pri(tdev, SN_ERR, "mem: get mem, id:%d is error\n",
			       memp.task_id);
			return -EFAULT;
		}

		ret = alloc_mem_ep(&addr, memp.size, memp.task_id, memp.fd,
				   tmem);

		if (ret) {
			sn_pri(tdev, SN_ERR, "mem: alloc memory failed.\n");
			return -EFAULT;
		}
		memp.phy_addr = addr;
		if (__copy_to_user(argp, &memp, sizeof(memp))) {
			sn_pri(tdev, SN_ERR,
			       "mem: get mem, copy_to_user failed, then free it\n");
			free_mem_ep(addr, memp.size, memp.task_id, tmem);
			return -EFAULT;
		}
		break;
	case SN_TRANX_MEM_FREE:
		if (copy_from_user(&memp, argp, sizeof(memp))) {
			sn_pri(tdev, SN_ERR,
			       "mem: free mem - copy_from_user failed.\n");
			return -EFAULT;
		}
		if (check_task_id(tmem, memp.task_id)) {
			sn_pri(tdev, SN_ERR, "mem: free mem,id:%d is error\n",
			       memp.task_id);
			return -EFAULT;
		}

		ret = free_mem_ep(memp.phy_addr, memp.size, memp.task_id, tmem);

		if (ret) {
			sn_pri(tdev, SN_ERR, "mem: free memory failed\n");
			return -EFAULT;
		}
		break;
	case SN_TRANX_MEM_GET_TASKID:
		id = get_task_id(tmem, filp);
		if (id == -1) {
			sn_pri(tdev, SN_ERR, "mem: can't get valid id\n");
			return -EFAULT;
		}
		sn_pri(tdev, SN_DBG, "mem: get task id:%d.\n", id);
		__put_user(id, (int *)argp);
		break;
	case SN_TRANX_MEM_FREE_TASKID:
		__get_user(id, (int *)argp);
		if (check_task_id(tmem, id)) {
			sn_pri(tdev, SN_DBG, "mem: free id, id:%d no memory left\n",
			       id);
		} else {
			free_task_id(tmem, id);
			sn_pri(tdev, SN_DBG, "mem: free task id:%d.\n", id);
		}
		break;
	default:
		sn_pri(tdev, SN_ERR, "mem: %s, cmd:0x%x is error.\n", __func__,
		       cmd);
		ret = -EINVAL;
	}

	return ret;
}
